import 'dart:async';

import 'package:bean_collector_annotation/bean_collector_annotation.dart';
import 'package:bean_collector_generator/src/bean_info.dart';
import 'package:build/build.dart';
import 'package:glob/glob.dart';
import 'package:source_gen/source_gen.dart';

/// @author newtab on 2021/7/6

const _outputExtensions = '.bean_collector.dart';

class BeanCollectorBuilder extends Builder {
  @override
  FutureOr<void> build(BuildStep buildStep) async {
    final resolver = buildStep.resolver;
    if (!await resolver.isLibrary(buildStep.inputId)) return;
    final lib = await buildStep.inputLibrary;
    final reader = LibraryReader(lib);
    String output;
    if (reader
        .annotatedWith(TypeChecker.fromRuntime(BeanCollectorBase))
        .isNotEmpty) {
      output = await _generate(reader, buildStep);
      await buildStep.writeAsString(
        buildStep.inputId.changeExtension(_outputExtensions),
        output,
      );
    }
  }

  @override
  Map<String, List<String>> get buildExtensions => {
        '.dart': [_outputExtensions]
      };

  _generate(LibraryReader reader, BuildStep buildStep) async {
    List<BeanInfo> beanCollectors = [];
    final pattern = '**/*.dart';
    final assetIds = await buildStep.findAssets(Glob(pattern)).toList()
      ..sort();
    final resolver = buildStep.resolver;
    for (var inputId in assetIds) {
      if (!await resolver.isLibrary(inputId)) continue;
      final lib = await resolver.libraryFor(inputId);
      final beanCollector =
          await _generateBeanInfo(LibraryReader(lib), inputId);
      if (beanCollector != null) {
        beanCollectors.add(beanCollector);
      }
    }
    return '''
$defaultFileHeader

${beanCollectors.map((e) => e.import).join('\n')}


class BeanCollector {

 static M fromJsonAsT<M>(dynamic json) {
    if (json is List) {
      return _getListChildType<M>(json);
    } else {
      return _fromJsonSingle<M>(json);
    }
  }

  static M _fromJsonSingle<M>(dynamic json) {

    String type = M.toString();

${beanCollectors.map((e) {
      return """
    if(type == (${e.name}).toString()){
      return ${e.name}().fromJson(json)  as M;
    }
    """;
    }).join('\n')}
    throw Exception("not found");
  }

  static M _getListChildType<M>(List<dynamic> data) {
  ${beanCollectors.map((e) {
      return """
    if(<${e.name}>[] is M){
      return data.map<${e.name}>((e) => ${e.name}().fromJson(e)).toList() as M;
    }
    """;
    }).join('\n')}
    throw Exception("not found");
  }
}
''';
  }

  _generateBeanInfo(LibraryReader library, AssetId inputId) {
    try {
      final annotatedElement =
          library.annotatedWith(TypeChecker.fromRuntime(BeanCollector)).first;
      final className = annotatedElement.element.displayName;
      final path = inputId.path;
      final package = inputId.package;
      final import =
          "import 'package:$package/${path.replaceFirst('lib/', '')}';";
      return BeanInfo(import, className);
    } on StateError catch (_) {
      return null;
    }
  }
}
